var valtio = (function (exports, react, proxyCompare, useSubscription) {
  'use strict';

  /*
  export {
    unstable_createMutableSource as createMutableSource,
    unstable_useMutableSource as useMutableSource,
  } from 'react'
  */
  var TARGET = Symbol();
  var createMutableSource = function createMutableSource(target, _getVersion) {
    var _ref;

    return _ref = {}, _ref[TARGET] = target, _ref;
  };
  var useMutableSource = function useMutableSource(source, getSnapshot, _subscribe) {
    var subscription = react.useMemo(function () {
      return {
        getCurrentValue: function getCurrentValue() {
          return getSnapshot(source[TARGET]);
        },
        subscribe: function subscribe(callback) {
          return _subscribe(source[TARGET], callback);
        }
      };
    }, [source, getSnapshot, _subscribe]);
    return useSubscription.useSubscription(subscription);
  };

  var MUTABLE_SOURCE = Symbol();
  var LISTENERS = Symbol();
  var SNAPSHOT = Symbol();

  var isObject = function isObject(x) {
    return typeof x === 'object' && x !== null;
  };

  var globalVersion = 0;
  var snapshotCache = new WeakMap();

  var createProxy = function createProxy(initialObject) {
    if (initialObject === void 0) {
      initialObject = {};
    }

    var version = globalVersion;
    var mutableSource;
    var listeners = new Set();

    var notifyUpdate = function notifyUpdate(nextVersion) {
      if (!nextVersion) {
        nextVersion = ++globalVersion;
      }

      if (version !== nextVersion) {
        version = nextVersion;
        listeners.forEach(function (listener) {
          return listener(nextVersion);
        });
      }
    };

    var emptyCopy = Array.isArray(initialObject) ? [] : Object.create(initialObject.constructor.prototype);
    var proxy = new Proxy(emptyCopy, {
      get: function get(target, prop, receiver) {
        if (prop === MUTABLE_SOURCE) {
          if (!mutableSource) {
            mutableSource = createMutableSource(receiver);
          }

          return mutableSource;
        }

        if (prop === LISTENERS) {
          return listeners;
        }

        if (prop === SNAPSHOT) {
          var cache = snapshotCache.get(receiver);

          if (cache && cache.version === version) {
            return cache.snapshot;
          }

          var snapshot = Object.create(target.constructor.prototype);
          snapshotCache.set(receiver, {
            version: version,
            snapshot: snapshot
          });
          Reflect.ownKeys(target).forEach(function (key) {
            var value = target[key];

            if (!isObject(value)) {
              snapshot[key] = value;
            } else if (value instanceof Promise) {
              Object.defineProperty(snapshot, key, {
                get: function get() {
                  throw value;
                }
              });
            } else {
              snapshot[key] = value[SNAPSHOT];
            }
          });
          return snapshot;
        }

        return target[prop];
      },
      deleteProperty: function deleteProperty(target, prop) {
        var value = target[prop];
        var childListeners = isObject(value) && value[LISTENERS];

        if (childListeners) {
          childListeners.delete(notifyUpdate);
        }

        delete target[prop];
        notifyUpdate();
        return true;
      },
      set: function set(target, prop, value, receiver) {
        var childListeners = isObject(target[prop]) && target[prop][LISTENERS];

        if (childListeners) {
          childListeners.delete(notifyUpdate);
        }

        if (!isObject(value)) {
          target[prop] = value;
        } else if (value instanceof Promise) {
          target[prop] = value.then(function (v) {
            receiver[prop] = v;
          });
        } else {
          var _getUntrackedObject;

          value = (_getUntrackedObject = proxyCompare.getUntrackedObject(value)) != null ? _getUntrackedObject : value;

          if (value[LISTENERS]) {
            target[prop] = value;
          } else {
            target[prop] = createProxy(value);
          }

          target[prop][LISTENERS].add(notifyUpdate);
        }

        notifyUpdate();
        return true;
      }
    });
    Reflect.ownKeys(initialObject).forEach(function (key) {
      proxy[key] = initialObject[key];
    });
    return proxy;
  };

  var subscribe = function subscribe(proxy, callback) {
    proxy[LISTENERS].add(callback);
    return function () {
      proxy[LISTENERS].delete(callback);
    };
  };

  var useProxy = function useProxy(proxy) {
    var affected = new WeakMap();
    var lastAffected = react.useRef();
    react.useEffect(function () {
      lastAffected.current = affected;
    });
    var getSnapshot = react.useMemo(function () {
      var prevSnapshot = null;
      var deepChangedCache = new WeakMap();
      return function (proxy) {
        var snapshot = proxy[SNAPSHOT];

        try {
          if (prevSnapshot !== null && lastAffected.current && !proxyCompare.isDeepChanged(prevSnapshot, snapshot, lastAffected.current, deepChangedCache)) {
            // not changed
            return prevSnapshot;
          }
        } catch (e) {// ignore and return new snapshot
        }

        return prevSnapshot = snapshot;
      };
    }, []);
    var snapshot = useMutableSource(proxy[MUTABLE_SOURCE], getSnapshot, subscribe);
    var proxyCache = react.useMemo(function () {
      return new WeakMap();
    }, []); // per-hook proxyCache

    return proxyCompare.createDeepProxy(snapshot, affected, proxyCache);
  };

  exports.proxy = createProxy;
  exports.subscribe = subscribe;
  exports.useProxy = useProxy;

  Object.defineProperty(exports, '__esModule', { value: true });

  return exports;

}({}, React, proxyCompare, useSubscription));
